#!/usr/bin/perl -w

use strict;
use GD;
use Getopt::Long;
use File::Basename;

my $revision_number = '$Revision: 1.8 $'; # '  $Date: 2004/08/28 04:24:53 $
$revision_number =~ s/[\$Revision\: | \$]//g;

my ($out_dir,$min_threshold,$max_threshold,$print_overall,$draw_lines,$draw_boxplots,$file_label);
my $sort_keys = 'name';
my %gene_colour = ();
my @gene_colours = ();
my $output_name = 'pathwayplots.html';
my $show_referer = '';

&GetOptions('out=s', \$out_dir,
	    'name=s', \$output_name,
	    'min=f', \$min_threshold,
	    'max=f', \$max_threshold,
	    'sort=s', \$sort_keys,
	    'single', \$print_overall,
	    'lines', \$draw_lines,
	    'box', \$draw_boxplots,
	    'refer=s', \$show_referer,
	    'label=s', \$file_label);


unless ($out_dir) {
    die "Please specify out directory via -out switch!\n";
}
unless (-d $out_dir) {
    die "Directory '$out_dir' doesn't exist!\n";
}
unless (-w $out_dir) {
    die "Can't write to directory '$out_dir'!\n";
}

my @sort_keys = ();
foreach (split /,/, $sort_keys) {
    # get rid of white space
    s/^\s*//;
    s/\s*$//;
    push @sort_keys, $_;
}

# get the max values
my $path_max = -9999;
my $path_min = 9999;
my $val_max = -9999;
my $val_min = 9999;
my $name_max = 0;
my %paths = ();
my %coords = ();
my %values = ();
my %genes = ();

my @labels = ();
my %labels = ();
my $legend_html = '';
if ($file_label) {
    foreach (split / \+\+ /, $file_label) {
	my ($file,$label) = split /=/;
	my $file_name = $file;
	$file_name =~ s/(.*)\.pathwayplot.*/$1/;
	$legend_html .= "$label: $file_name<BR>\n";
#	warn "#KH: $file, $label";
	$labels{$file} = $label;
	push @labels, $label;
    }
}

# all the remaining arguments should be filenames
my @files = @ARGV;

if (@files == 1) {
    $print_overall = 1;
} else {
    $print_overall = 0 unless (defined $print_overall);
}

my $file = '';

foreach my $file (@files) {
    open (IN, $file)
	or die "Can't read from file '$file': $!\n";
    
    # read in all genes found on the array
    while (<IN>) {
	chomp;
	my ($gene,$val,$path_num,$path_name) = split /\t/, $_, 4;

	$genes{$gene}++;
	
	# get min and max of value / path number
	if ($val > $val_max) {
	    $val_max = $val;
	} 
	if ($val < $val_min) {
	    $val_min = $val;
	}
	if ($path_num > $path_max) {
	    $path_max = $path_num;
	} 
	if ($path_num < $path_min) {
	    $path_min = $path_num;
	}
	
	# get longest name
	if ( (length $path_name) > $name_max) {
	    $name_max = length $path_name;
	}
	
	# keep hash of path name/number match:
	if (defined $paths{$path_name}{$file}) {
	    if ($paths{$path_name}{$file} ne $path_num) {
		warn "$path_name has two numbers: '$paths{$path_name}{$file}' and '$path_num'!\n";
	    }
	} else {
	    $paths{$path_name}{$file} = $path_num;
	}

	# collect coordinate pairings:
	push @{$coords{$file}{$path_name}}, "$val|$gene";
	push @{$values{$file}{$path_name}}, $val;
    }
    close IN;
}

my $max_num = 0;

# sort pathways		    	    
my %sort = ();
my %vals = ();
foreach my $path (keys %paths) {
    
    # key1: number of values
    my %path_num = ();
    foreach my $file (@files) {
	my $num;
	if (defined $values{$file}{$path} ) {
	    $num = scalar @{$values{$file}{$path}};
	    $vals{'dots'}{$file}{$path} = $num;
	    if ($num > $max_num) {
		$max_num = $num;
	    }
	    push @{$path_num{$path}}, $num;
	}
    }
    my $num_med = &median(@{$path_num{$path}});
    push @{$sort{'dots'}{$num_med}}, $path;    
    $vals{'dots'}{$path} = $num_med;   


    # key3: median
    my %path_med = ();
    foreach my $file (@files) {
	my @tmp = ();
	if (defined $coords{$file}{$path} ) {
	    foreach (@{$coords{$file}{$path}}) {
		my ($val,$gene) = split /\|/, $_;
		push @tmp, $val;
	    }
	    my $med = &median(@tmp);
	    push @{$path_med{$path}}, $med;
	}
    }
    my $med_med = &median(@{$path_med{$path}});
    push @{$sort{'median'}{$med_med}}, $path;
    $vals{'medians'}{$path} = $med_med;   
}

my $key = shift @sort_keys;
unless (defined $sort{$key}) {
    die "# ERROR: key '$key' not known!\n" unless ($key eq 'name');
}

my @order = ();

if ($key eq 'name') {
    @order = sort keys %paths;
} else {
    my @tmp2 = ();
    @tmp2 = sort { $b <=> $a } keys %{$sort{$key}};

    foreach (@tmp2) {
	push @order, @{$sort{$key}{$_}};
    }
}

if ($print_overall) {

    foreach my $file (@files) {
	
	my $out_html = basename $file;
	if (@files == 1) {
	    $out_html = "$out_dir/$output_name";
	} else {
	    $out_html .= '.html';
	    $out_html = "$out_dir/$out_html";
	}

	open (HTML, ">$out_html")
	    or die "Can't write to out file '$out_html': $!\n";
	my $out_img = $out_html;
	$out_img =~ s/\.html$/\.png/;
	my $out_img_html = basename $out_img;
	
	open (OUT, ">$out_img")
	    or die "Can't write to $out_img: $!\n";

	my $map = "<MAP NAME=\"map\">\n";
	
	&print_diagram_overall($file,\$map);
	
	$map .= "</MAP>\n";
	
	#write html page around it:
	print HTML "<HTML>
<HEAD>
<TITLE>
pathway expression plotter
</TITLE>
</HEAD>
<BODY BGCOLOR=WHITE>
<H1>Pathway Expression Plotter</H1>
$show_referer
<IMG SRC=\"$out_img_html\" USEMAP=\"#map\">
$map
</BODY>
</HTML>
";
	
	close HTML;
    }
} else {

    my $out_html = basename $file;
    $out_html = "$out_dir/$output_name";
    
    open (HTML, ">$out_html")
	or die "Can't write to out file '$out_html': $!\n";

	#write html page around it:
	print HTML "<HTML>
<HEAD>
<TITLE>
pathway expression plotter
</TITLE>
</HEAD>
<BODY BGCOLOR=WHITE>
<H1>Pathway Expression Plotter</H1>
$show_referer
One graph per path with genes plotted according to their fold-changes in each condition.
<P>
<B>Legend</B>:<BR>
$legend_html
<BR>
";

    foreach my $path (@order) {

	my $image_link = &print_diagram_pathway($path);
	print HTML "$image_link\n";
    }

    print HTML "</BODY>
</HTML>
";
    
    close HTML;
    
}


sub print_diagram_overall {

    my $file = shift;
    my $map_ref = shift;
    my $map = $$map_ref;

    # create a new image    
    my $space_left = 20;
    my $space_right = 20;
    my $space_top = 50;
    my $space_bottom = 20;
    my $gene_font = gdTinyFont; 
    my $legend_font = gdSmallFont; 
    my $text_font = gdMediumBoldFont;

    my $spot_diameter = 7;
    my $spot_spacing = 10;
    my $rad = sprintf "%.0f", ($spot_diameter/2);

    my $label_width = $legend_font->width * $name_max + 20;
    my $graph_width = 1000;
    my $graph_height = ($spot_diameter + $spot_spacing) * ( (scalar keys %paths) + 1);

    my $left_overhang = $legend_font->width * length($max_num) + 5;


    my $all_x = $space_left + $space_right + $graph_width + $label_width;
    
    my $all_y = $space_top + $space_bottom + $graph_height;
    
    my $im = new GD::Image($all_x,$all_y);
    
    # allocate some colors
    my $white = $im->colorAllocate(255,255,255);
    my $black = $im->colorAllocate(0,0,0);
    my $red = $im->colorAllocate(255,0,0);
    my $blue = $im->colorAllocate(0,0,255);
    my $green = $im->colorAllocate(0,255,0);
    my $grey = $im->colorAllocate(224,224,224);
    my $midgrey = $im->colorAllocate(199,199,199);
    my $darkgrey = $im->colorAllocate(99,99,99);
    my $orange = $im->colorAllocate(255,165,0);
    my $yellow = $im->colorAllocate(255,255,0);
    my $cyan = $im->colorAllocate(0,255,255);
    my $magenta = $im->colorAllocate(255,0,255);
    my $brown = $im->colorAllocate(103,67,0);
    my $green_yellow = $im->colorAllocate(159,211,0);
    my $pink = $im->colorAllocate(255,174,185);
    my $navy = $im->colorAllocate(0,0,142);
    my $gold = $im->colorAllocate(255,215,0);
    my $turquoise = $im->colorAllocate(72,209,204);
    my $violet = $im->colorAllocate(148,0,211);
    my $maroon = $im->colorAllocate(103,7,72);
    my $thistle = $im->colorAllocate(146,62,112);
    my $khaki = $im->colorAllocate(189,167,107);
    my $sienna = $im->colorAllocate(142,107,35);
    my $coral = $im->colorAllocate(248,137,117);
    my $olive = $im->colorAllocate(59,94,43);
    my $teal = $im->colorAllocate(0,128,128);
    my $purple = $im->colorAllocate(128,0,128);
    my $dark_green = $im->colorAllocate(0,83,0);
    my $silver = $im->colorAllocate(192,192,192);
    my $plum = $im->colorAllocate(76,46,87);
    my $orchid = $im->colorAllocate(218,107,212);

    # define coordinates of origin
    my $startx = $space_left;
    my $starty = $space_top + $graph_height;
    my ($x1,$x2,$y1,$y2,$xt);


    # get tic marks for x-axis:

    my ($x_min,$x_max,$front,$end);

    if (defined $min_threshold) {
	$x_min = $min_threshold;
    } else {
	# get the next lowest border (to one decimal)
	$x_min = sprintf "%.1f", $val_min;
	$x_min =~ /(\d+)\.(\d)/;
	$front = $1;
	$end = $2;
	if ($x_min < 0) {
	    if ($end == 9) {
		$end = 0;
		$front++;
	    } else {
		$end++;
	    }
	    $x_min = -1 * ($front + $end/10);
	} else {
	    if ($end == 0) {
		$end = 9;
		$front--;
	    } else {
		$end--;
	    }
	    $x_min = $front + $end/10;
	}
    }

    if (defined $max_threshold) {
	$x_max = $max_threshold;
    } else {
	$x_max = sprintf "%.1f", $val_max;
	$x_max =~ /(\d+)\.(\d)/;
	$front = $1;
	$end = $2;
	if ($x_max < 0) {
	    if ($end == 0) {
		$end = 9;
		$front--;
	    } else {
		$end--;
	    }
	    $x_max = -1 * ($front + $end/10);
	} else {
	    if ($end == 9) {
		$end = 0;
		$front++;
	    } else {
		$end++;
	    }
	    $x_max = $front + $end/10;
	}
    }

    
    my $scale_x = $graph_width / abs($x_max - $x_min);

    my @borders = ($x_min);

    for (my $i = $x_min+0.1; $i < $x_max; $i += 0.1) {
	my $k = sprintf "%.1f", $i;
	my $l = $k * 10;
	if ( ($l % 5) == 0) {
	    push @borders, $k;
	}
    }
    push @borders, $x_max;

    # print name of file
    $xt = $all_x/2 - ($legend_font->width * length(basename $file))/2;
    $im->string($legend_font,$xt,10,(basename $file),$black);

    # draw x-tics and labels
    $y1 = $starty - 2;
    $y2 = $starty + 2;

    my $y1_up = $space_top - 2;
    my $y2_up = $space_top + 2;

    for my $i (0..$#borders) {
	$x1 = $startx + ($borders[$i] - $x_min) * $scale_x;	
	# draw tic
	$im->line($x1,$y1,$x1,$y2,$black);
	# print label
	my $text = $borders[$i];
	$xt = $x1 - ($legend_font->width * length($text))/2;
	$im->string($legend_font,$xt,$y2,$text,$black);

	# upper tics
	$im->line($x1,$y1_up,$x1,$y2_up,$black);
	$im->string($legend_font,$xt,$y1_up-13,$text,$black);
    }

    # draw boxes and dots for each pathway
    $y1 = $starty - $spot_spacing;
    my @colours = ($grey,$white);
    my $endx = $startx + $graph_width + $label_width - 15;

    my $box_x = $startx - $left_overhang;


    # print differently shaded boxes as background
    for my $i (0..$#order) {
	
	my $colour = shift @colours;
	push @colours, $colour;

	$y1 -= ($spot_spacing + $spot_diameter);
	$y2 = $y1 + ($spot_spacing + $spot_diameter);

	# print box:
	$im->filledRectangle($box_x,$y1,$endx,$y2,$colour);
    }


    # draw grid
    $y1 = $starty - 2;
    for my $i (0..$#borders) {
	$x1 = $startx + ($borders[$i] - $x_min) * $scale_x;	
	$im->line($x1,$y1,$x1,$y2_up,$midgrey);
    }


    $y1 = $starty - $spot_spacing;
    for my $i (0..$#order) {
	# print pathname
	my $path = $order[$i];

	$y1 -= ($spot_spacing + $spot_diameter);
	$y2 = $y1 + ($spot_spacing + $spot_diameter);

	$xt = $startx + $graph_width + 5;
	$im->string($legend_font,$xt,$y1+2,$path,$black);

	# print number of dots
	my $num = $vals{'dots'}{$path};
	$xt = $startx - 2 - ($legend_font->width * length($num));
	$im->string($legend_font,$xt,$y1+2,$num,$black);

	# get all values for this pathway
	next unless (defined $values{$file}{$path});

	my @values = @{$coords{$file}{$path}};

	# print each
	my %printed = ();
	foreach (@values) {
	    my ($val,$gene) = split /\|/, $_, 2;
	    $x1 = $startx + (sprintf "%.0f", ($val - $borders[0]) * $scale_x);
	    my $y = $y1 + (sprintf "%.0f", ($y2 - $y1)/2);

	    # check for duplicate dots with the same expression value
	    if ($printed{$x1}) {
		$y = $printed{$x1} - (sprintf "%.0f", $spot_diameter/2);
	    } else {
		$printed{$x1} = $y;
	    }

	    $im->arc($x1,$y,$spot_diameter,$spot_diameter,0,360,$blue);
	    for (my $i = $spot_diameter-1; $i > 0; $i--) {
		$im->arc($x1,$y,$i,$i,0,360,$blue);
	    }
#	    $im->fill($x1,$y,$blue);
#	    $im->fillToBorder($x1,$y,$blue,$blue); # this doesn't work because it leaves some spots (overlapping with others) white

	    $val = sprintf "%.2f", $val;
	    $val =~ s/(.*\d+)\.0+$/$1/;
	    $val =~ s/(.*\d+\.\d+)0+$/$1/;

	    $map .= "<AREA SHAPE=CIRCLE COORDS=\"$x1,$y,$rad\" TITLE=\"$gene, $val\" HREF=\"#$gene,$val\" TARGET=viewer  ALT=\"$gene, $val\"  OnMouseOut=\"window.status=''; return true\"  OnMouseOver=\"window.status='$gene, $val'; return true\">\n";
	}

	next unless (defined $values{$file}{$path});

	@values = @{$values{$file}{$path}};

	unless (@values) {
	    warn "No values for $file, $path!\n";
	    next;
	}

	# draw box-plot:
	my ($min,$quant,$md,$quant2,$max) = &quantile(0.25,\@values);
	$x1 = $startx + ($quant - $borders[0]) * $scale_x;
	$x2 = $startx + ($quant2 - $borders[0]) * $scale_x;
	$im->rectangle($x1,$y1+2,$x2,$y2-2,$black);
	$x1 = $startx + ($md - $borders[0]) * $scale_x;
	$im->line($x1,$y1+2,$x1,$y2-2,$red);	
    }

    # draw axes:
    $y2 = $space_top;
    $x2 = $startx + $graph_width; 
    $im->line($startx,$starty+2,$startx,$y2,$black);
    $im->line($box_x,$starty,$x2+10,$starty,$black);
    $im->line($startx+$graph_width,$starty+2,$startx+$graph_width,$y2,$black);
    $im->line($box_x,$y2,$x2+10,$y2,$black);


    # make sure we are writing to a binary stream
    binmode OUT;
    
    # Convert the image to PNG and print it on standard output
    print OUT $im->png;
    close OUT;
}



sub print_diagram_pathway {

    my $path = shift;

    my $out_img = $path;
    $out_img =~ s/\W/_/g;
    $out_img = "$out_dir/$out_img.png";
    my $out_img_html = basename $out_img;
    
    open (OUT, ">$out_img")
	or die "Can't write to $out_img: $!\n";
    
    
    # determine the number of files with that path
    my @files = sort keys %{$paths{$path}};
    
    # get longest name
    my $name_max = 0;
    foreach (@files) {
	if ( (length basename $_) > $name_max) {
	    $name_max = length basename $_;
	}
    }		    

    if (@labels) {
	$name_max = 0;
	foreach (@labels) {
	    if (length($_) > $name_max) {
		$name_max = length($_);
	    }
	}
    }
    
    # create a new image    
    my $space_left = 20;
    my $space_right = 20;
    my $space_top = 50;
    my $space_bottom = 40;
    my $gene_font = gdTinyFont; 
    my $legend_font = gdSmallFont; 
    my $text_font = gdMediumBoldFont;

    my $spot_diameter = 7;
    my $spot_spacing = 10;
    my $rad = sprintf "%.0f", ($spot_diameter/2);

    my $label_width = $legend_font->width * $name_max + 20;
    my $graph_width = 1000;
    my $graph_height = ($spot_diameter + $spot_spacing) * ( (scalar @files) + 1);

    my $left_overhang = $legend_font->width * length($max_num) + 5;


    my $all_x = $space_left + $space_right + $graph_width + $label_width;

    my $all_y = $space_top + $space_bottom + $graph_height;

    my $im = new GD::Image($all_x,$all_y);

    # allocate some colors
    my $white = $im->colorAllocate(255,255,255);
    my $black = $im->colorAllocate(0,0,0);
    my $red = $im->colorAllocate(255,0,0);
    my $blue = $im->colorAllocate(0,0,255);
    my $green = $im->colorAllocate(0,255,0);
    my $grey = $im->colorAllocate(224,224,224);
    my $midgrey = $im->colorAllocate(199,199,199);
    my $darkgrey = $im->colorAllocate(99,99,99);
    my $orange = $im->colorAllocate(255,165,0);
    my $yellow = $im->colorAllocate(255,255,0);
    my $cyan = $im->colorAllocate(0,255,255);
    my $magenta = $im->colorAllocate(255,0,255);
    my $brown = $im->colorAllocate(103,67,0);
    my $green_yellow = $im->colorAllocate(159,211,0);
    my $pink = $im->colorAllocate(255,174,185);
    my $navy = $im->colorAllocate(0,0,142);
    my $gold = $im->colorAllocate(255,215,0);
    my $violet = $im->colorAllocate(148,0,211);
    my $maroon = $im->colorAllocate(103,7,72);
    my $thistle = $im->colorAllocate(146,62,112);
    my $khaki = $im->colorAllocate(189,167,107);
    my $sienna = $im->colorAllocate(142,107,35);
    my $coral = $im->colorAllocate(248,137,117);
    my $olive = $im->colorAllocate(59,94,43);
    my $teal = $im->colorAllocate(0,128,128);
    my $purple = $im->colorAllocate(128,0,128);
    my $dark_green = $im->colorAllocate(0,83,0);
    my $silver = $im->colorAllocate(192,192,192);
    my $plum = $im->colorAllocate(76,46,87);
    my $orchid = $im->colorAllocate(218,107,212);
    my $col1 = $im->colorAllocate(0,255,102);
    my $col2 = $im->colorAllocate(255,204,153);
    my $col3 = $im->colorAllocate(153,51,0);
    my $col4 = $im->colorAllocate(119,119,119);
    my $col5 = $im->colorAllocate(0,102,255);

    @gene_colours = ($red,
		     $blue,
		     $green,
		     $orange,
		     $gold,
		     $cyan,
		     $magenta,
		     $brown,
		     $green_yellow,
		     $pink,
		     $navy,
		     $yellow,
		     $violet,
		     $maroon,
		     $thistle,
		     $khaki,
		     $coral,
		     $olive,
		     $teal,
		     $purple,
		     $dark_green,
		     $sienna,
		     $plum,
		     $orchid,
		     $col1,
		     $col2,
		     $col3,
		     $col4,
		     $col5,
		     );
    
    # assign colours to genes (sorted by occurence in descending order)
    my %colours_done = ();

    # define coordinates of origin
    my $startx = $space_left;
    my $starty = $space_top + $graph_height;
    my ($x1,$x2,$y1,$y2,$xt);


    # get tic marks for x-axis:

    my ($x_min,$x_max,$front,$end);

    if (defined $min_threshold) {
	$x_min = $min_threshold;
    } else {
	# get the next lowest border (to one decimal)
	$x_min = sprintf "%.1f", $val_min;
	$x_min =~ /(\d+)\.(\d)/;
	$front = $1;
	$end = $2;
	if ($x_min < 0) {
	    if ($end == 9) {
		$end = 0;
		$front++;
	    } else {
		$end++;
	    }
	    $x_min = -1 * ($front + $end/10);
	} else {
	    if ($end == 0) {
		$end = 9;
		$front--;
	    } else {
		$end--;
	    }
	    $x_min = $front + $end/10;
	}
    }

    if (defined $max_threshold) {
	$x_max = $max_threshold;
    } else {
	$x_max = sprintf "%.1f", $val_max;
	$x_max =~ /(\d+)\.(\d)/;
	$front = $1;
	$end = $2;
	if ($x_max < 0) {
	    if ($end == 0) {
		$end = 9;
		$front--;
	    } else {
		$end--;
	    }
	    $x_max = -1 * ($front + $end/10);
	} else {
	    if ($end == 9) {
		$end = 0;
		$front++;
	    } else {
		$end++;
	    }
	    $x_max = $front + $end/10;
	}
    }


    my $scale_x = $graph_width / abs($x_max - $x_min);

    my @borders = ($x_min);

    for (my $i = $x_min+0.1; $i < $x_max; $i += 0.1) {
	my $k = sprintf "%.1f", $i;
	my $l = $k * 10;
	if ( ($l % 5) == 0) {
	    push @borders, $k;
	}
    }
    push @borders, $x_max;


    # print name of pathway
    $xt = $all_x/2 - ($legend_font->width * length(basename $path))/2;
    $im->string($legend_font,$xt,10,(basename $path),$black);

    # draw x-tics and labels
    $y1 = $starty - 2;
    $y2 = $starty + 2;

    my $y1_up = $space_top - 2;
    my $y2_up = $space_top + 2;

    for my $i (0..$#borders) {
	$x1 = $startx + ($borders[$i] - $x_min) * $scale_x;	
	# draw tic
	$im->line($x1,$y1,$x1,$y2,$black);
	# print label
	my $text = $borders[$i];
	$xt = $x1 - ($legend_font->width * length($text))/2;
	$im->string($legend_font,$xt,$y2,$text,$black);

	# upper tics
	$im->line($x1,$y1_up,$x1,$y2_up,$black);
	$im->string($legend_font,$xt,$y1_up-13,$text,$black);
    }

    # draw boxes and dots for each pathway
    $y1 = $starty - $spot_spacing;
    my @colours = ($grey,$white);
    my $endx = $startx + $graph_width + $label_width - 15;

    my $box_x = $startx - $left_overhang;


    # print differently shaded boxes as background
    for my $i (0..$#files) {

	my $colour = shift @colours;
	push @colours, $colour;

	$y1 -= ($spot_spacing + $spot_diameter);
	$y2 = $y1 + ($spot_spacing + $spot_diameter);

	# print box:
	$im->filledRectangle($box_x,$y1,$endx,$y2,$colour);
    }


    # draw grid
    $y1 = $starty - 2;
    for my $i (0..$#borders) {
	$x1 = $startx + ($borders[$i] - $x_min) * $scale_x;	
	$im->line($x1,$y1,$x1,$y2_up,$midgrey);
    }
    
    my $map_name = $path;
    $map_name =~ s/\W/_/g;
    my $map = "<MAP NAME=\"$map_name\">\n";;
    $y1 = $starty - $spot_spacing;
    
    my %dots = ();

    for my $file (sort @files) {

	# get all values for this pathway
	next unless (defined $values{$file}{$path});

	my @values = @{$coords{$file}{$path}};
	next unless (@values);

	$y1 -= ($spot_spacing + $spot_diameter);
	$y2 = $y1 + ($spot_spacing + $spot_diameter);

	# print file name
	$xt = $startx + $graph_width + 5;
	my $text = basename $file;
	if (defined $labels{$text}) {
	    $text = $labels{$text};
	}
	$im->string($legend_font,$xt,$y1+2,$text,$black);

	# print number of dots
	my $num = $vals{'dots'}{$file}{$path};
	$xt = $startx - 2 - ($legend_font->width * length($num));
	$im->string($legend_font,$xt,$y1+2,$num,$black);

	# print each
	my %printed = ();
	foreach (@values) {
	    my ($val,$gene) = split /\|/, $_, 2;
	    $x1 = $startx + (sprintf "%.0f", ($val - $borders[0]) * $scale_x);
	    my $y = $y1 + (sprintf "%.0f", ($y2 - $y1)/2);

	    # check for duplicate dots with the same expression value
	    if ($printed{$x1}) {
		$y = $printed{$x1} - (sprintf "%.0f", $spot_diameter/2);
	    } else {
		$printed{$x1} = $y;
	    }

	    # take note of coordinates for later (first we'll draw the connection lines)
	    if (defined $dots{$gene}{$y}{$x1}) {
		warn "#WARNING: multiple values for '$path', '$gene', '$y', '$x1': '$val' and '$dots{$gene}{$y}{$x1}'!\n";
	    }

	    $dots{$gene}{$y}{$x1} = $val;

	}
    }


    # draw dots and connecting lines:

    my $gene_xt1 = 10;
    my $gene_xt2 = 10;

    my @gene_colours2 = @gene_colours;
    %colours_done = ();
#    foreach my $gene (sort { $genes{$b} <=> $genes{$a} } keys %dots) {
    foreach my $gene (sort keys %dots) {
	my @y = sort { $b <=> $a } keys %{$dots{$gene}};
	my $colour = '';
	if (@y < 2) {
	    $colour = $black;
	} else {
	    my $counter = 0;
	    my $found = 0;
	    while ($counter <= $#gene_colours2) {		    
		$colour = shift @gene_colours2;
		push @gene_colours2, $colour;
		unless (defined $colours_done{$colour}) {
		    $found = 1;
		    last;
		}
		$counter++;
	    }
	    unless ($found) {
		$colour = $gene_colours2[int(rand $#gene_colours2)];
	    }
	    $gene_colour{$gene} = $colour;
	    $colours_done{$colour}++;
	}

	# draw name in chosen colour:
	if (@y < 2) {
	    $im->string($gene_font,$gene_xt1,$all_y-10,$gene,$colour);
	    $gene_xt1 += $gene_font->width * length($gene) + 5;
	} else {
	    $im->string($gene_font,$gene_xt2,$all_y-20,$gene,$colour);
	    $gene_xt2 += $gene_font->width * length($gene) + 5;
	}

	for (my $i = 0; $i <= $#y; $i++) {
	    my $y1 = $y[$i];
	    if ( scalar (keys %{$dots{$gene}{$y1}}) > 1) {
		warn "#WARNING: more than one x coordinate for '$path', '$gene', y1 ('$y1')!\n";
	    }
	    foreach my $x1 (keys %{$dots{$gene}{$y1}}) {
		if (defined $draw_lines
		    and
		    $draw_lines > 0
		    and
		    $i < $#y) {
		    my $y2 = $y[$i+1];
		    if ( scalar (keys %{$dots{$gene}{$y2}}) > 1) {
			warn "#WARNING: more than one x coordinate for '$path', '$gene', '$y2'!\n";
		    }
		    foreach my $x2 (keys %{$dots{$gene}{$y2}}) {
			$im->line($x1,$y1,$x2,$y2,$colour);
		    }
		}
	    
		&draw_spot($x1,$y1,$spot_diameter,$spot_diameter,$colour,$im);
		my $val = $dots{$gene}{$y1}{$x1};
		$val = sprintf "%.2f", $val;
		$val =~ s/(.*\d+)\.0+$/$1/;
		$val =~ s/(.*\d+\.\d+)0+$/$1/;
		
		$map .= "<AREA SHAPE=CIRCLE COORDS=\"$x1,$y1,$rad\" TITLE=\"$gene, $val\" HREF=\"#$gene,$val\" TARGET=viewer  ALT=\"$gene, $val\"  OnMouseOut=\"window.status=''; return true\"  OnMouseOver=\"window.status='$gene, $val'; return true\">\n";
	    }
	}
    }

    # draw boxplots:
    $y1 = $starty - $spot_spacing;
    
    for my $file (sort @files) {
	
	next unless (defined $draw_boxplots
		     and
		     $draw_boxplots > 0);

	next unless (defined $values{$file}{$path});

	my @values = @{$values{$file}{$path}};

	unless (@values) {
	    warn "No values for $file, $path!\n";
	    next;
	}

	$y1 -= ($spot_spacing + $spot_diameter);
	$y2 = $y1 + ($spot_spacing + $spot_diameter);

	my ($min,$quant,$md,$quant2,$max) = &quantile(0.25,\@values);
	$x1 = $startx + ($quant - $borders[0]) * $scale_x;
	$x2 = $startx + ($quant2 - $borders[0]) * $scale_x;
	$im->rectangle($x1,$y1+2,$x2,$y2-2,$black);
	$x1 = $startx + ($md - $borders[0]) * $scale_x;
	$im->line($x1,$y1+2,$x1,$y2-2,$red);	

    }    

    # draw axes:
    $y2 = $space_top;
    $x2 = $startx + $graph_width; 
    $im->line($startx,$starty+2,$startx,$y2,$black);
    $im->line($box_x,$starty,$x2+10,$starty,$black);
    $im->line($startx+$graph_width,$starty+2,$startx+$graph_width,$y2,$black);
    $im->line($box_x,$y2,$x2+10,$y2,$black);

    $map .= "</MAP>\n";
    
    # make sure we are writing to a binary stream
    binmode OUT;

    # Convert the image to PNG and print it on standard output
    print OUT $im->png;
    close OUT;
    
    return "<IMG SRC=\"$out_img_html\" USEMAP=\"#$map_name\">\n$map";
}


    
sub draw_spot {
    my ($x,$y,$diameter1,$diameter2,$colour,$im) = @_;
    $im->arc($x,$y,$diameter1,$diameter2,0,360,$colour);
    for (my $i = $diameter1-1; $i > 0; $i--) {
	$im->arc($x,$y,$i,$i,0,360,$colour);
    }
#	    $im->fill($x1,$y,$blue);
#	    $im->fillToBorder($x1,$y,$blue,$blue); # this doesn't work because it leaves some spots (overlapping with others) white
}


sub quantile {
    
    # details from the R help(quantile) text:
    # 
    #   `quantile(x,p)' as a function of `p' linearly interpolates the
    #   points ( (i-1)/(n-1), ox[i] ), where `ox <- sort(x)' and `n <-
    #   length(x)'.
    #  
    #   This gives `quantile(x, p) == (1-f)*ox[i] + f*ox[i+1]', where `r
    #   <- 1 + (n-1)*p', `i <- floor(r)', `f <- r - i' and `ox[n+1] :=
    #   ox[n]'.

    my $percentile = shift;
    my $ref = shift;
    my @data = @$ref;
    
    my @order = (sort { $a <=> $b } @data);
    my $n = scalar @data;
    my $rank = 1 + ($n - 1) * $percentile;
    my $i = &floor($rank);
    my $num1 = $order[$i-1];
    my $num2 = ($i > $#order ? $order[-1] : $order[$i]);
    my $fraction = $rank - $i;
    my $quant = (1 - $fraction) * $num1 + $fraction * $num2;

    # get the inverse quant as well
    $percentile = 1 - $percentile;
    $rank = 1 + ($n - 1) * $percentile;
    $i = &floor($rank);
    $num1 = $order[$i-1];
    $num2 = ($i > $#order ? $order[-1] : $order[$i]);
    $fraction = $rank - $i;
    my $quant2 = (1 - $fraction) * $num1 + $fraction * $num2;
    
    # get the median:    
    my $md = &median(@order);

    # return min, first quant, median, second quant, max:
    return $order[0],$quant,$md,$quant2,$order[-1];
}

sub floor {

    # get the floor of a number 
    # (greatest integer that is lower or equal to number)

    my $num = shift;
    my $floor;
    if ($num >= 0) {
	$floor = int($num);
    } else {
	$floor = int($num);
	unless ($floor == $num) {
	    $floor -= 1;
	}
    }
    return $floor;
}



sub median {
    my @order = sort { $a <=> $b } @_;
    my $md = '';

    unless (@order) {
	warn "No input!\n";
	return '';
    }

    if ($#order % 2 == 0){
        $md = $order[$#order / 2];
    } else {
	my $index1 = int($#order / 2);
	my $index2 = int($#order / 2) + 1;
	my $low = $order[$index1];
	my $high = $order[$index2];
	$md = $low + ($high - $low)/2;
    }
    return $md;
}

    

__END__


 
# sort pathways
my @order = ();
my @out = ();
my @sort_keys = ('name');
my $key = shift @sort_keys;
my @vals = reverse sort keys %{$sort{$key}};
my %unsorted = ();
my $i = 0;
foreach my $key2 (@vals) {
    my @tmp = @{$sort{$key}{$key2}};
    if (@tmp > 1) {
	$unsorted{$i}++;
    }
    push @{$out[$i]}, [ @tmp ];
    $i++;
}

while (%unsorted) {
    my @out2 = @out;
    @out = ();
    $key = shift @sort_keys;
    @vals = reverse sort keys %{$sort{$key}};
    my %unsorted2;
    for my $i (0..$#out2) {
	unless ($unsorted{$i}) {
	    push @out, [ @{$out2[$i]} ];
	} else {
	    my %vals = ();
	    foreach (@{$out2[$i]}) {
		$vals{$_}++;
	    }
	    foreach my $key2 (@vals) {
		my @tmp = @{$sort{$key}{$key2}};
		
		foreach (@tmp) {
		    if ($vals{$_}) {
			push @out, [ $_ ];
		    }
		} else {
